import scala.collection.mutable.ArrayBuffer
import java.io.{File, FileInputStream, FileOutputStream}
import scala.actors.Actor._
import scala.util.Sorting

object driver {
  def main(args : Array[String]) : Unit = {
    if (args.size != 1) {
      println("Usage: ...")
      return
    }
    new driver (new File (args(0)))
  }
  
  class TupleIterator(size : Int, maxValue : Int) extends Iterator[List[Int]] {
    private val current = Array.fill(size)(0)
    private var nextExists = true
    def hasNext = nextExists
    def next : List[Int] = {
      val res = current.toList
      for (i <- 0 until size)
        if (current(i) < maxValue) {
          current(i) = current(i) + 1
          return res
        } else {
          current(i) = 0
        }
      nextExists = false
      res
    }
  }

  private val InvariantRegex = """(?<!free)\s+(invariant[^;]+;)""".r
  private val ImplRegex = """(procedure|implementation)\s+([^(\s]+)\s*\(""".r
  private val UnwindRegex = """kUnwind\s+[0-9]+;""".r
  private val Boogie = "c:\\Users\\philipp\\Desktop\\boogie\\boogie-sources-svn\\Binaries\\Boogie.exe"
  private val BoogieSuccess = """Boogie program verifier finished with [0-9]+ verified, 0 errors""".r
  private val BoogieTimeout = """Boogie program verifier finished with.*out""".r

  private val TimeLimit = 200
  private val KLimit = 3

  object CheckResult extends Enumeration {
    val Correct, CounterExample, Timeout = Value
  }

  def measureTime[A](comp : => A) : (A, Long) = {
    val timeBefore = System.currentTimeMillis
    val res = comp
    val timeAfter = System.currentTimeMillis
    (res, timeAfter - timeBefore)
  }

  def printTime[A](comp : => A) : A = {
    val (res, time) = measureTime(comp)
    println("" + time + "ms")
    res
  }
}

class driver(bplFile : File) {

  import driver._
  
  println("Using time limit " + TimeLimit)
  println("Using maximum k=" + KLimit)

  private val bplContents = {
    println("Reading " + bplFile + " ...")
    val stream = new FileInputStream (bplFile)
    val buf = new StringBuffer
    var c = stream.read
    while (c != -1) {
      buf append c.toChar
      c = stream.read
    }
    buf.toString    
  }
 
  //////////////////////////////////////////////////////////////////////////////
  
  private val invariantPos = 
    (new Iterator[(Int, Int)] {
      private val matchIt = InvariantRegex findAllIn (bplContents)
      def hasNext = matchIt.hasNext
      def next : (Int, Int) = {
        matchIt.next
        (matchIt start 1, matchIt end 1)
      }
    }).toList
  
  private val unwindPos = 
    (new Iterator[(Int, Int)] {
      private val matchIt = UnwindRegex findAllIn (bplContents)
      def hasNext = matchIt.hasNext
      def next : (Int, Int) = {
        matchIt.next
        (matchIt start, matchIt end)
      }
    }).toList
  
  private val implPos = 
    (new Iterator[(Int, Int)] {
      private val matchIt = ImplRegex findAllIn (bplContents)
      def hasNext = matchIt.hasNext
      def next : (Int, Int) = {
        matchIt.next
        (matchIt start 2, matchIt end 2)
      }
    }).toList

  // figure out the contained invariants and
  // unwind statements per procedure

  private val procInvariantPos = {
    val rawPos =
      for (((nameBegin, nameEnd), (nextImpl, _)) <-
           (implPos zip ((implPos drop 1) ++ List((Int.MaxValue, 0)))))
      yield (bplContents.substring(nameBegin, nameEnd)
             -> ((for ((ib, ie) <- invariantPos;
                       if (ib >= nameEnd && ie <= nextImpl)) yield (ib, ie)),
                 (for ((ib, ie) <- unwindPos;
                       if (ib >= nameEnd && ie <= nextImpl)) yield (ib, ie)))
             )
    for (n <- (for ((a, _) <- rawPos) yield a).distinct) yield (n ->
      ((for ((a, (b, _)) <- rawPos; if (n == a); x <- b) yield x),
       (for ((a, (_, b)) <- rawPos; if (n == a); x <- b) yield x))
    )
  }

  /////////////////////////////////////////////////////////////////////////////

  println("Checking individual procedures ...")

  for ((procName, (invariantPos, unwindPos)) <- procInvariantPos;
       if (!invariantPos.isEmpty)) {

  println("Procedure: " + procName + ", "
          + unwindPos.size + " loops, "
          + invariantPos.size + " invariants")

  def genVariant(enabledInvs : List[Boolean],
                 unwindValues : List[Int],
                 outfile : File) = {
    val invReplacement =
      for ((enabled, bounds@(start, end)) <- enabledInvs zip invariantPos) yield (
        (bounds -> (if (enabled) bplContents.substring(start, end) else "")))
    val unwindReplacement =
      for ((k, bounds) <- unwindValues zip unwindPos) yield (
        (bounds -> ("kUnwind " + k + ";")))

    val replacements = (invReplacement ++ unwindReplacement).toArray
    Sorting.stableSort[((Int, Int), String), Int](
                       replacements, { case ((start, _), _) => start })

    val stream = new FileOutputStream (outfile)
    Console.withOut(stream) {
      var currentPos = 0
      for (((start, end), str) <- replacements) {
        print(bplContents.substring(currentPos, start))
        print(str)
        currentPos = end
      }
      print(bplContents.substring(currentPos, bplContents.size))
    }
    stream.close
  }

  //////////////////////////////////////////////////////////////////////////////

  def checkAndPrint(enabledInvs : List[Boolean],
                    unwindValues : List[Int],
                    options : List[String]) : Boolean =
    check(enabledInvs, unwindValues, options) match {
      case CheckResult.Correct        => { print("_"); true }
      case CheckResult.Timeout        => { print("T"); false }
      case CheckResult.CounterExample => { print("C"); false }
    }
  
  def check(enabledInvs : List[Boolean],
            unwindValues : List[Int],
            options : List[String]) : CheckResult.Value = {
    val outfile = File.createTempFile("lessInvs", ".bpl")
    try {
      genVariant(enabledInvs, unwindValues, outfile)

      val args = List(Boogie) ++ options ++
                 List("/proc:" + procName, "/timeLimit:" + TimeLimit, outfile.getAbsolutePath)

      val p = Runtime.getRuntime.exec(args.toArray)
      val selfActor = self
                                    
      // check the stdout stream for counterexamples
      val stdout = p.getInputStream
      val stdoutActor = actor {
        val buf = new StringBuffer
        var c = stdout.read
        while (c != -1) {
          buf append c.toChar
          c = stdout.read
        }

        selfActor ! (
          if ((BoogieTimeout findFirstIn buf.toString).isDefined)
            CheckResult.Timeout
          else if ((BoogieSuccess findFirstIn buf.toString).isDefined)
            CheckResult.Correct
          else
            CheckResult.CounterExample
        )
      }

      // record the stderr stream, in case we want to show it later
      val stderr = p.getErrorStream
      actor {
        val buf = new StringBuffer
        var c = stderr.read
        while (c != -1) {
          buf append c.toChar
          c = stderr.read
        }
      
        stdoutActor ! buf.toString
      }
    
      p.waitFor
      receive {
        case x : CheckResult.Value => x
      }
    } finally {
      outfile.delete
    }
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  // Greedily find out which of the invariants are not needed even with the
  // standard Boogie loop rule
  
  val zeroUnwindValues = for (_ <- unwindPos) yield 0

  val neededInvs = {
    print("  Checking required invariants w/o  k-induction ")
    
    val invs = Array.fill(invariantPos.size)(true)

    var changed = true

    while (changed) {
      changed = false
      for (i <- 0 until invariantPos.size) {
        if (invs(i)) {
          invs(i) = false
          if (!checkAndPrint(invs.toList, zeroUnwindValues, List()))
            invs(i) = true
          else
	    changed = true
        } else {
	  print("_")
        }
      }
      if (changed) {
        println
        print("                                                ")
      }
    }

    invs.toList
//    List.fill(invariantPos.size)(true)  // ignore this information in this case
  }
  
  println(", " + (neededInvs count (true == _)) + " invariants")
  
  print("  Time w/o k-induction: ")
  printTime { check(neededInvs, zeroUnwindValues, List()) }

  //////////////////////////////////////////////////////////////////////////////
  
  var bestUnwindValues : List[Int] = null
  var bestNeededInvNum : Int = 1000000
  var bestUnneededInvs : Set[Int] = null

  var best2UnwindValues : List[Int] = null
  var best2RemovableInvNum : Int = -1
  var best2MaxUnneededInvSets : Set[Set[Int]] = null

  for (unwindValues <- new TupleIterator(unwindPos.size, KLimit)) {

  println(" Testing k = " + unwindValues)

  // check whether the program can be verified using those ks at all

  print("  Checking with k-induction and all invariants: ")
  if (printTime(check(neededInvs, unwindValues, List("/kInduction:0"))) != CheckResult.Correct) {
    println("  Program cannot be verified using this k, even with all invariants")
  } else {

  //////////////////////////////////////////////////////////////////////////////

  // Find out which invariants can individually be removed with the k-induction
  // rule
  
  val kIndUnneededInvs = {
    print("  Checking required invariants with k-induction ")
    
    val invs = Array.fill(invariantPos.size)(true)
    for (i <- 0 until invariantPos.size;
         if ({
               if (neededInvs(i)) {
                 checkAndPrint(neededInvs.updated(i, false), unwindValues,
                               List("/kInduction:0"))
               } else {
                 print("_")
                 false
               }
              }))
      yield i
  }.toSet
  
  val kIndNeededInvNum =
    (neededInvs count (true == _)) - kIndUnneededInvs.size
  println(", " + kIndNeededInvNum + " invariants")
  
  //////////////////////////////////////////////////////////////////////////////
  
  // Greedily check whether we can combine any of the unnecessary invariants
  
  val maxUnneededInvSets : Set[Set[Int]] = {
    print("  Greedily searching for maximum sets of unneeded invariants ")
    
    val res = new ArrayBuffer[Set[Int]]
    var usedInvs = Set[Int]()
    
    while (!(kIndUnneededInvs.toSet subsetOf usedInvs)) {
      print(".")
      
      val seed = kIndUnneededInvs.find({ case x => !(usedInvs contains x) }).get
      var currentInvs = neededInvs.updated(seed, false)
      
      usedInvs += seed
      
      for (i <- kIndUnneededInvs; if (!(usedInvs contains i))) {
        val newInvs = currentInvs.updated(i, false)
        if (check(newInvs, unwindValues, List("/kInduction:0")) == CheckResult.Correct)
          currentInvs = newInvs
      }
      
      val selectedInvs =
        (for (i <- kIndUnneededInvs; if (!currentInvs(i))) yield i).toSet
      usedInvs ++= selectedInvs
      res += selectedInvs
    }
    
    if (res.isEmpty)
      Set(Set())
    else
      res.toSet
  }
  
  println
  println("  Maximum sets: " + maxUnneededInvSets)

  if (kIndNeededInvNum < bestNeededInvNum) {
    bestNeededInvNum = kIndNeededInvNum
    bestUnwindValues = unwindValues
    bestUnneededInvs = kIndUnneededInvs
  }
  val maxRemovable = (for (s <- maxUnneededInvSets) yield s.size).max
  if (maxRemovable > best2RemovableInvNum) {
    best2RemovableInvNum = maxRemovable
    best2UnwindValues = unwindValues
    best2MaxUnneededInvSets = maxUnneededInvSets
  }
  }}

  if (bestUnneededInvs != null) {
    println(" Best k for individual removal: " + bestUnwindValues)
    for (r <- bestUnneededInvs) {
      print("  Time with k-induction, removing " + r + ": ")
      val enabledInvs = neededInvs.updated(r, false)
      val res = printTime {
        check(enabledInvs, bestUnwindValues, List("/kInduction:0"))
      }
      assert(res == CheckResult.Correct)
    }
  }
  if (best2MaxUnneededInvSets != null) {
    println(" Best k for collective removal: " + best2UnwindValues)
    for (r <- best2MaxUnneededInvSets) {
      print("  Time with k-induction, removing " + r + ": ")
      val enabledInvs = List.tabulate(invariantPos.size) {
        case i => neededInvs(i) && !(r contains i)
      }
      val res = printTime {
        check(enabledInvs, best2UnwindValues, List("/kInduction:0"))
      }
      assert(res == CheckResult.Correct)
    }
  }
  }
}
